/* 
 *    Programmed By: Mohammed Isam Mohammed [mohammed_isam1984@yahoo.com]
 *    Copyright 2014, 2015, 2016 (c)
 * 
 *    file: calc.c
 *    This file is part of tcalc.
 *
 *    tcalc is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    tcalc is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with tcalc.  If not, see <http://www.gnu.org/licenses/>.
 */    

#include <unistd.h>
#include "defs.h"

//extern char *strndup(const char *S, size_t SIZE);

char res2[64];

/*
 * Preprocess subformula, by doing error checking, before
 * passing it to process_sub_formula_f() to do the actual
 * calculations. Called from calc().
 */
int preprocess_subformula(char *subf, char *ERROR, double *res, struct custom_func_s *active_func)
{
  *ERROR = 0;
  if(!strrchr(subf, ')'))
  {
    fprintf(stderr, "Error: Missing ')'\n");
    *ERROR = 1; return 0;
  }
  int j;
  int open_braces = 1, close_braces = 0;
  for(j = 1; j < strlen(subf); j++)
  {
    if(subf[j] == ')') close_braces++;
    else if(subf[j] == '(') open_braces++;
    if(open_braces == close_braces) break;
  }
  if(close_braces < open_braces)
  {
    fprintf(stderr, "Error: Missing ')'\n");
    *ERROR = 1; return 0;
  }
  if(close_braces > open_braces)
  {
    fprintf(stderr, "Error: Missing '('\n");
    *ERROR = 1; return 0;
  }
  *res = process_sub_formula_f(subf, j, ERROR, active_func);
  return j;
}

/*
 * Extracts a numerical, passed as a string in 'tmp', then
 * applies the 'op' operator on it and the result in 'res',
 * then passes the result back in 'res'.
 */
void do_op(char *tmp, char op, double *res, char *ERROR)
{
	double result_f = *res;
	double tmp_res = 0;
	if(tmp[0] == 'b' || tmp[0] == 'B')
	  tmp_res = binary_to_decimal(tmp+1, ERROR);
	else if((tmp[0] == '0' && tmp[1] == 'X') ||
	        (tmp[0] == '0' && tmp[1] == 'x'))
	  tmp_res = hex_to_decimal(tmp, ERROR);
	else if(tmp[0] == '0' && tmp[1] != '.')
	  tmp_res = octal_to_decimal(tmp, ERROR);
	else tmp_res = atof(tmp);
	if(*ERROR) return;

	switch(op)
	{
	  case '+': result_f += tmp_res; break;
	  case '-': result_f -= tmp_res; break;
	  case '*': result_f *= tmp_res; break;
	  case '/': result_f /= tmp_res; break;
	  case '%': result_f = (int)result_f%(int)tmp_res; break;
	  //case '^': result_f = (double)((int)result_f^(int)tmp_res); break;
	  case '&': result_f = (int)result_f & (int)tmp_res; break;
	  case '|': result_f = (int)result_f | (int)tmp_res; break;
	  /* FIXME: Check if this works! */
	  case '!': if(tmp_res) result_f = __do_factorial((long)tmp_res); break;
	  case '=':
	  case '.': result_f = tmp_res; break;
	}
	*res = result_f;
}

void __do_op_float(char op, double *result_f, double tmp_res)
{
  switch(op)
  {
    case '+': (*result_f) += tmp_res; break;
    case '-': (*result_f) -= tmp_res; break;
    case '*': (*result_f) *= tmp_res; break;
    case '/': (*result_f) /= tmp_res; break;
    case '%': (*result_f) = (int)(*result_f)%(int)tmp_res; break;
    //case '^': (*result_f) = (int)(*result_f)^(int)tmp_res; break;
    case '&': (*result_f) = (int)(*result_f) & (int)tmp_res; break;
    case '|': (*result_f) = (int)(*result_f) | (int)tmp_res; break;
    case '=':
    case '.':
      (*result_f) = tmp_res; break;
  }
}

void __do_op_long(char op, long *result, double *result_f, double tmp_res)
{
  switch(op)
  {
    case '+': (*result) += tmp_res; break;
    case '-': (*result) -= tmp_res; break;
    case '*': (*result) *= tmp_res; break;
    case '/':
      (*result_f) = (*result); number_mode = NUMBER_FLOAT;
      (*result_f) /= tmp_res;
      break;
    case '%': (*result) = (*result)%(int)tmp_res; break;
    //case '^': (*result) = (*result)^(int)tmp_res; break;
    case '&': (*result) = (*result) & (int)tmp_res; break;
    case '|': (*result) = (*result) | (int)tmp_res; break;
    case '=':
    case '.':
      (*result) = tmp_res; break;
  }
}

/*
 * List the operators in the passed formula. Called from
 * reorder_ops() for collecting info about ops before
 * working on them.
 */
int collect_ops(char *formula, int len, char *ops, 
		int op_index[], char op_precedence[],
		int *op_count, int max_count)
{
  int i, count = 0;
  for(i = 0; i < len; i++)
  {
    if(is_op(formula[i]) && formula[i] != '\t' && formula[i] != ' ')
    {
      ops[count] = formula[i];
      if(formula[i] == '(' || formula[i] == ')') op_precedence[count] = 6;
      else if(formula[i] == '!') op_precedence[count] = 5;
      else if(formula[i] == '*' || formula[i] == '/') op_precedence[count] = 4;
      else if(formula[i] == '%' || formula[i] == '^') op_precedence[count] = 4;
      else if(formula[i] == '+' || formula[i] == '-') op_precedence[count] = 3;
      else if(formula[i] == '&' || formula[i] == '|') op_precedence[count] = 2;
      else op_precedence[count] = 0;
      op_index[count++] = i;
      if(count == max_count) break;
    }
  }
  *op_count = count;
  if(max_count == count) return 1;
  return 0;
}

#define MAX_OPS		32

/*
 * Reorders the operators in the formula by inserting braces
 * around higher order ops appearing after lower order ops,
 * thus effectively forcing functions that will later process
 * this formula to respect the order of the braces.
 */
int reorder_ops(char *formula, int len)
{
  char ops[MAX_OPS];
  char op_precedence[MAX_OPS];
  int op_index[MAX_OPS];
  int op_count = 0;
  if(!formula || !len) return -1;
  collect_ops(formula, len, ops, op_index, op_precedence, &op_count, MAX_OPS);
  if(op_count <= 1) return -2;
  int i;
  /*
  for(i = 0; i < op_count; i++)
  {
    printf("%c   %4d   %2d\n", ops[i], op_index[i], op_precedence[i]);
  }
  */
  for(i = 1; i < op_count; i++)
  {
    if((op_precedence[i] > op_precedence[i-1]) && op_precedence[i] != 6)
    {
      int j;
      for(j = i+1; j < op_count; j++)
      {
	if(op_precedence[j] < op_precedence[i])
	{
	  formula[len+2] = '\0';
	  int k;
	  for(k = len+2; k > op_index[j]+1; k--) formula[k] = formula[k-2];
	  formula[k] = ')';
	  for(--k; k > op_index[i-1]+1; k--) formula[k] = formula[k-1];
	  formula[k] = '(';
	  for(k = i; k < j; k++) op_index[k]++;
	  for(; k < op_count; k++) op_index[k] += 2;
	  len += 2;
	  break;
	}
      }
    //printf("[%s]\n", formula);
    }
  }
  /*
  for(i = 0; i < op_count; i++)
  {
    printf("%c   %4d   %2d\n", ops[i], op_index[i], op_precedence[i]);
  }
  printf("[%s]\n", formula);
  */
  return 0;
}

/*
 * Does the donkey work of processing a formula for calcuation.
 * Called from input_free_mode() and input_file_mode().
 */
double calc(char *orig_formula, char *ERROR, char *COMMENT, struct custom_func_s *active_func)
{
  int i = 0, k;
  char *tmp;
  long result = 0;
  double result_f = 0;
  char op = '.';
  number_mode = 0;
  *ERROR = 0;
  char special_number = 0;	/* flag to denote if we are in the middle of binary or hex num */
  
  char formula[2048];
  strcpy(formula, orig_formula);
  int FORMULA_LEN = strlen(formula);
  reorder_ops(formula, FORMULA_LEN);
  FORMULA_LEN = strlen(formula);
  
  int active_var = -1;
  int active_var_scope = GLOBAL_SCOPE;
  
  for(k = 0; k <= strlen(formula); k++)
  {
    if(is_whitespace(formula[k])) continue;
    if(formula[k] == '{' || formula[k] == '}')
    {
      *ERROR = 3; break;
    }
    else if(formula[k] == '(')
    {
      double tmp_res;
      int j = preprocess_subformula(formula+k, ERROR, &tmp_res, active_func);
      if(*ERROR) break;
      if(strchr(formula+k, '.'))
      {
	if(number_mode == NUMBER_INT)
	{ result_f = result; number_mode = NUMBER_FLOAT; }
      }

      if(number_mode == NUMBER_FLOAT)
      {
	__do_op_float(op, &result_f, tmp_res);
      }
      else
      {
	__do_op_long(op, &result, &result_f, tmp_res);
      }
      k += j;
      op = formula[k];
      i = k+1;
      if(special_number) special_number = 0;
    }
    /////////////////////////////////
    /////////////////////////////////
    /////////////////////////////////
    else if(formula[k] == '+' || formula[k] == '-' ||
	    formula[k] == '*' || formula[k] == '/' ||
	    formula[k] == '%' || formula[k] == '\0'||
	    /* formula[k] == '^' || */ formula[k] == '&' ||
	    formula[k] == '|' || formula[k] == '\n')
    {
      tmp = strndup(formula+i, k-i);
      if(number_mode == NUMBER_FLOAT)
      {
calc_float:
        do_op(tmp, op, &result_f, ERROR);
      }
      else
      {
	if(strchr(tmp, '.'))
	{
	  number_mode = NUMBER_FLOAT;
	  result_f = result; result = 0;
	  goto calc_float;
	}
	else
	{
	  number_mode = NUMBER_INT;
	  result_f = result;
          do_op(tmp, op, &result_f, ERROR);
	  result = (long)result_f;
	  if(op == '/')
	    number_mode = NUMBER_FLOAT;
	}
      }
      if(*ERROR) break;
      op = formula[k];
      i = k+1;
      if(special_number) special_number = 0;
    }
    ///////////////////////////////////////
    //process math functions
    ///////////////////////////////////////
    else if(formula[k] >= 'a' && formula[k] <= 'z')
    {
      /* Maybe we are misreading a binary or decimal number? */
      if(special_number) continue;
      if(k < FORMULA_LEN-1 && formula[k] == 'b'&&
	(formula[k+1] == '1' || formula[k+1] == '0')) { special_number = 1; continue; }
      if(k && formula[k-1] == '0' && (formula[k] == 'x' || formula[k] == 'X'))
      { special_number = 1; continue; }
      /* Maybe this is a math function name or a variable? */
      char *ob = strchr(formula+k, '(');
      char *cb = strchr(formula+k, ')');
      char *eq = strchr(formula+k, '=');
      double r;
      if((!ob && !cb) ||
	 (eq && eq < ob && ob < cb))
      {
	int len = 0;
	while(formula[k+len] >= 'a' && formula[k+len] <= 'z') len++;
	if(!len) break;
	int res = -1;
	int index = -1;
	if((res = is_var(formula+k, len, &index, active_func)))
	{
	  /* seek for variable assignment */
	  if(eq && (eq >= (formula+k+len)))
	  {
	    active_var = index;
	    if(res == 1) active_var_scope = FUNCTION_SCOPE(1);
	    else if (res == 2) active_var_scope = GLOBAL_SCOPE;
	    k = eq-formula;
	    r = result_f;
	    goto proc;
	  }
	  
	  r = get_var(formula+k, len, active_func, ERROR);
	  if(*ERROR) break;
	  k += len;
	  goto proc;
	}
	printf("Variable ");
	for(index = 0; index < len; index++) putchar(formula[k+index]);
	printf(" is not defined\n");
	*ERROR = 1; break;
      }
      if(!ob || !cb)
      {
	/* Nope. The user messed up. Tell him Jim! */
	if(!ob) printf("Error: missing '('\n");
	else    printf("Error: missing ')'\n");
	result = -1; result_f = -1; *ERROR = 1; break;
      }
      r = process_math_func(formula+k, ob-(formula+k), cb-(formula+k), ERROR, active_func);
      if(*ERROR) break;
      k = cb-(formula+k)+1;
proc:
      if(number_mode == NUMBER_INT)
      { result_f = result; }
      number_mode = NUMBER_FLOAT;
      __do_op_float(op, &result_f, r);
      op = formula[k];
      i = k+1;
    }
  }//end for
  
  if(number_mode == NUMBER_INT)
    result_f = result;
  /* do we need to store the value in a variable? */
  if(active_var >= 0)
  {
    /* We don't want the caller to output the result,
     * so set error.
     */
    (*ERROR) = 2;
    if(active_var_scope == GLOBAL_SCOPE)
      __do_op_float('.', &global_var_value[active_var], result_f);
    else
      __do_op_float('.', &active_func->var_value[active_var], result_f);
  }
  return result_f;
}


/*
 * Output the final result from any given calculation.
 * If the number_mode is decimal, will use 'result',
 * otherwise (i.e. is float), will use 'result_f'.
 */
void output_res(int output_format, double result_f, long int result)
{
  char error = 0;
  char *res3 = res2;
  if(output_format == DECIMAL_FORMAT || output_format == ALL_FORMATS)
  {
    if(decimal_point)
    {
      char tmp2[20], tmp3[10];
      tmp2[0] = '%';
      tmp2[1] = '.';
      tmp2[2] = '\0';
      sprintf(tmp3, "%d", decimal_point);
      strcat(tmp2, tmp3);
      strcat(tmp2, "f\n");
      if(number_mode == NUMBER_INT)
	result_f = result;
      printf(tmp2, result_f);
    }
    else
    {
      if(number_mode == NUMBER_INT) printf("%+ld (%lu)\n", result, result);
      else printf("%.0f\n", result_f);
    }
  }
  if(output_format == OCTAL_FORMAT || output_format == ALL_FORMATS)
  {
    if(number_mode == NUMBER_INT)
      printf("0%lo\n", (unsigned long)result);
    else printf("0%lo\n", (unsigned long)result_f);
  }
  if(output_format == HEX_FORMAT || output_format == ALL_FORMATS)
  {
    if(number_mode == NUMBER_INT)
    {
      decimal_to_hex((unsigned long)result, &res3, &error);
      if(*res2) { printf("%s\n", res2); }
    }
    else
    {
      decimal_to_hex((unsigned long)result_f, &res3, &error);
      if(*res2) { printf("%s\n", res2); }
    }
  }
  if(output_format == BINARY_FORMAT || output_format == ALL_FORMATS)
  {
    if(number_mode == NUMBER_INT)
    {
      decimal_to_binary((unsigned long)result, &res3, &error);
      if(*res2) { printf("b%s\n", res2); }
    }
    else
    {
      decimal_to_binary((unsigned long)result_f, &res3, &error);
      if(*res2) { printf("b%s\n", res2); }
    }
  }
}
