/* 
 *    Programmed By: Mohammed Isam Mohammed [mohammed_isam1984@yahoo.com]
 *    Copyright 2014, 2015, 2016, 2020 (c)
 * 
 *    file: defs.h
 *    This file is part of tcalc.
 *
 *    tcalc is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    tcalc is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with tcalc.  If not, see <http://www.gnu.org/licenses/>.
 */    

#ifndef _DEFS_H
#define _DEFS_H

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <asm/types.h>
#include <limits.h>
#include <string.h>
#include <math.h>


#ifndef __NUMBER_MODES
#define __NUMBER_MODES
#define NUMBER_INT	1
#define NUMBER_FLOAT	2
#endif /* __NUMBER_MODES */

#ifndef __FORMATS
#define __FORMATS
#define DECIMAL_FORMAT	1
#define HEX_FORMAT	2
#define BINARY_FORMAT	3
#define OCTAL_FORMAT	4
#define ALL_FORMATS	5
#endif /* __FORMATS */

//for custom functions
#ifndef __CUSTOM_FUNCS
#define __CUSTOM_FUNCS
#define MAX_FUNCTIONS		20
#define MAX_FUNCTION_NAME	10
#define MAX_VARIABLES		10
#define MAX_VARIABLE_LEN	10
#endif /* __CUSTOM_FUNCS */

#define MAX_DECIMAL_POINT	10

struct custom_func_s
{
  char custom_func_name[MAX_FUNCTION_NAME];
  char *func_body;
  char *func_args;
  int func_args_count;
  int func_body_len;	/* actual used space */
  int func_body_res;	/* reserved space */
  char var_name[MAX_VARIABLES][MAX_VARIABLE_LEN];
  double var_value[MAX_VARIABLES];
  int total_vars;
  struct custom_func_s *next, *prev;
};

/* defined in main.c */
extern struct custom_func_s custom_func_head;
extern int input_format, output_format;
extern int number_mode;
extern int decimal_point;
extern const char tcalc_ver[];
extern double process_sub_formula_f(char* subformula, int end, char* error, struct custom_func_s *active_func);
extern char *trim(char *n, char KEEP_NEWLINE);	/* main.c */
extern void print_mult_table(char *n, int from, int to, int step);
extern int QUIET;
extern int SCOPE;
#define GLOBAL_SCOPE		(0)
#define FILE_SCOPE		(-1)
#define FUNCTION_SCOPE(x)	(x)

/* number format functions */
/* defined in numbers.c    */
extern unsigned long hex_to_decimal(char *hex, char *error);
extern unsigned long binary_to_decimal(char *binary, char *error);
extern unsigned long octal_to_decimal(char *octal, char *error);
extern void decimal_to_hex(unsigned long decimal, char **res, char *error);
extern void decimal_to_binary(unsigned long decimal, char **res, char *error);
extern void convert_numbers(char *input);
extern int authenticate_hex(char *hex, char *error);
extern int authenticate_octal(char *octal, char *error);
extern int authenticate_binary(char *binary, char *error);
extern char is_hex_digit(char hex_digit);
extern char is_op(char op);

/* custom functions routines */
/* defined in custom_func.c  */
extern int total_custom_funcs;
extern void print_custom_funcs();
extern int process_custom_func(struct custom_func_s *func, char* args, double* res);
//extern void delete_custom_func(int func_no);
extern void delete_custom_func_by_name(char *func_name);
int is_custom_func(char* func, int len1);
extern double global_vars[MAX_VARIABLES];
extern int total_global_vars;
void delete_var_by_name(char* var, struct custom_func_s *active_func);
void delete_var(int var_no, struct custom_func_s *active_func);
void print_vars();
extern char global_var_name[MAX_VARIABLES][MAX_VARIABLE_LEN];
extern double global_var_value[MAX_VARIABLES];
extern int total_global_vars;
#define MAX_GLOBAL_VARS			10
struct custom_func_s *create_custom_func(char* name, int name_len, int argc);
void delete_custom_func(struct custom_func_s *func);
int add_function_args(struct custom_func_s *func, char *args, int len);
#define FUNCTION_BODY_BLOCK_SIZE	512
int is_var(char *var, int len, int *index, struct custom_func_s *active_func);
double get_var(char* var, int len, struct custom_func_s *active_func, char *error);

/* math functions routines */
/* defined in math.c       */
extern double process_math_func(char* func, int ob_index, int cb_index,
				char* error, struct custom_func_s *active_func);
extern void print_func_descriptions();
extern int total_funcs;
extern char *func_name[];
int is_reserved_func(char* name, int len1);

/* input in free mode      */
/* defined in input_free.c */
extern void input_free_mode();
extern void show_help();
extern int change_number_format(char *formula);
extern int change_output_format(char *formula);
extern void input_convert_numbers(char *formula);
extern void input_mult_table(char *formula);
long unsigned int __do_factorial(long unsigned int n);
long unsigned int input_do_factorial(char* formula);
void __input_file(char *formula, int request);

/* input from file mode    */
/* defined in input_file.c */
extern void input_from_file_mode(FILE* file, char list_only);

/* calculations routines */
/* defined in calc.c     */
extern double calc(char* orig_formula, char* ERROR, char* COMMENT, struct custom_func_s *active_func);
extern void output_res(int output_format, double result_f, long int result);
extern void do_op(char *tmp, char op, double *res, char *ERROR);
extern void __do_op_float(char op, double *result_f, double tmp_res);
extern void __do_op_long(char op, long *result, double *result_f, double tmp_res);
int preprocess_subformula(char *subf, char *ERROR, double *res, struct custom_func_s *active_func);

/* defined in proc_input.c */
extern int proc_input(char* formula, struct custom_func_s *active_func);

static inline char is_whitespace(char c)
{
  if(c == ' ' || c == '\t' || c == '\n') return 1;
  return 0;
}

static inline void remove_trailing_newline(char *str)
{
  int len = strlen(str);
  if(str[len-1] == '\n')
    str[len-1] = '\0';
}

#endif
