/* 
 *    Programmed By: Mohammed Isam Mohammed [mohammed_isam1984@yahoo.com]
 *    Copyright 2014, 2015 (c)
 * 
 *    file: input_file.c
 *    This file is part of tcalc.
 *
 *    tcalc is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    tcalc is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with tcalc.  If not, see <http://www.gnu.org/licenses/>.
 */    

#include "defs.h"


#define FREE_AND_RETURN(f)		\
{					\
  free(f);				\
  return;				\
}

#define EXIT_ON_FILE_ERROR(f)		\
{					\
  printf("Error reading from file.\n");	\
  FREE_AND_RETURN(f)			\
}

#define EXIT_ON_ERROR()			\
{					\
  printf("Error reading from file.\n");	\
  return;				\
}

#define ERROR_CONTINUE(s)	{fprintf(stderr, s); continue;}

/*
 * Reads a formula file passed as FILE *.
 * The flag determins if we are to process the file or not:
 * 0 - Process the file
 * 1 - List the file only (something like 'cat file')
 * 2 - List the functions defined within
 */
void input_from_file_mode(FILE *file, char list_only)
{
  long file_size;
  char *file_content;
  if(fseek(file, 0, SEEK_END) == -1) EXIT_ON_ERROR()
  file_size = ftell(file);
  if(fseek(file, 0, SEEK_SET) == -1) EXIT_ON_ERROR()
  file_content = (char *)malloc(file_size+1);
  if(!file_content)
  {
    printf("Error: Insufficient memory\n");
    return;
  }
  if(fread(file_content, 1, file_size, file) != file_size) 
  {
    //free(file_content);
    EXIT_ON_FILE_ERROR(file_content)
  }
  /* make sure the file content is NULL-terminated for later processing */
  file_content[file_size] = '\0';
  
  char LIST_FUNCS = 0;
  if(list_only == 1)
  {
    printf("%s\n", file_content);
    FREE_AND_RETURN(file_content)
  }
  else if(list_only == 2) LIST_FUNCS = 1;
  
  SCOPE = FILE_SCOPE;
  char *formula;
  formula = strtok(file_content, "\n");
  goto process;
  
  while((formula = strtok(NULL, "\n")))
  {
process:
	  if(0) ;	//to shut up gcc
	  char COMMENT = 0;
	  //int i;
	  char *s = trim(formula, 1);
	  strcpy(formula, s);
	  if(formula[0] == '#') { COMMENT = 1; continue; }
	  
	  if(LIST_FUNCS) goto skip;
	  
	  int p = proc_input(formula, NULL);
	  if(p > 0) break;	/* means there was an error */
	  else if(!p) continue;	/* means everything was processed right */
	  /* NOTE: negative return value from proc_input() means it couldn't
	   * process the directive. See if we can do it here.*/
skip:
	  if(strncmp(formula, "func", 4) == 0)
	  {
	    int l = 4;
	    while(is_whitespace(formula[l])) l++;
	    if(l == 4 || formula[l] < 'a' || formula[l] > 'z')
	    {
	      printf("Error: Missing or Invalid function name\n");
	      FREE_AND_RETURN(file_content)
	    }
	    int l2 = l+1;
	    int len = strlen(formula);
	    while(l2 < len && formula[l2] >= 'a' && formula[l2] <= 'z') l2++;
	    
	    int j;
	    char ERROR = 0;
	    struct custom_func_s *new_func = 0;
	    
	    if(LIST_FUNCS)
	    {
	      char c = formula[l2];
	      formula[l2] = '\0';
	      printf("%s", formula+l);
	      formula[l2] = c;
	    }
	    else
	    {
	      if(is_reserved_func(formula+l, l2-l)) FREE_AND_RETURN(file_content)
	      if(is_custom_func(formula+l, l2-l)) FREE_AND_RETURN(file_content)
	      new_func = create_custom_func(formula+l, l2-l, 0);
	      if(!new_func) FREE_AND_RETURN(file_content)
	    }
	    /* extract function arguments, if any */
	    l = 0;
	    int l3 = 0;
	    for(j = l2; j < len; j++)
	    {
	      if(formula[j] == '(') l = j+1;
	      if(formula[j] == ')') l3 = j-1;
	    }
	    if(l)
	    {
	      if(!l3)
	      {
		printf("Error: Missing ')'\n");
		FREE_AND_RETURN(file_content)
	      }
	      if(LIST_FUNCS)
	      {
		char c = formula[l3+1];
		formula[l3+1] = '\0';
		printf("(%s)\n", formula+l);
		formula[l3+1] = c;
	      }
	      else
	      {
		if(!add_function_args(new_func, formula+l, l3-l+1))
		{
		  printf("Error: Insufficient memory to save function args\n");
		  delete_custom_func(new_func);
		  FREE_AND_RETURN(file_content)
		}
	      }
	    }
	    
	    if(!LIST_FUNCS)
	    {
	      new_func->func_body = (char *)malloc(FUNCTION_BODY_BLOCK_SIZE);
	      if(!new_func->func_body)
	      {
		printf("Error: Insufficient memory to save function body\n");
		delete_custom_func(new_func);
		FREE_AND_RETURN(file_content)
	      }
	      new_func->func_body[0] = '\0';
	      new_func->func_body_res = FUNCTION_BODY_BLOCK_SIZE;
	    }
	    
	    while(!strchr(formula, '{'))
	      formula = strtok(NULL, "\n");
	    if(formula == NULL)
	    {
	      printf("Error: couldn't find '{' until end of file\n");
	      delete_custom_func(new_func);
	      FREE_AND_RETURN(file_content)
	    }
	    formula = strtok(NULL, "\n");
	    
	    while(!strchr(formula, '}'))
	    {
	      if(formula == NULL)
	      {
		printf("Error: couldn't find '}' until end of file\n");
		delete_custom_func(new_func);
		FREE_AND_RETURN(file_content)
	      }
	      if(!LIST_FUNCS)
	      {
		//printf("line 193\n");
		new_func->func_body_len += strlen(formula)+2;
		if(new_func->func_body_len >= new_func->func_body_res)
		{
		  new_func->func_body_res = (int)(new_func->func_body_len/FUNCTION_BODY_BLOCK_SIZE)*FUNCTION_BODY_BLOCK_SIZE;
		  if(new_func->func_body_len%FUNCTION_BODY_BLOCK_SIZE)
		    new_func->func_body_res += FUNCTION_BODY_BLOCK_SIZE;
		  new_func->func_body = (char *)realloc(new_func->func_body,
							new_func->func_body_res);
	          if(!new_func->func_body)
		  {
		    printf("Error: Insufficient memory to save function body\n");
		    delete_custom_func(new_func);
		    FREE_AND_RETURN(file_content)
		  }
		}
		strcat(new_func->func_body, formula);
		strcat(new_func->func_body, "\n");
	      }
	      formula = strtok(NULL, "\n");
	    }
	    if(ERROR) FREE_AND_RETURN(file_content)
	    //if(!LIST_FUNCS) total_custom_funcs++;
	    continue;
	    //goto process;
	  }
	  //////////////////////////////////////
	  //////////////////////////////////////
	  if(LIST_FUNCS) continue;
	  
	  char ERROR = 0;
	  long result = 0;
	  double result_f = 0;
	  result_f = calc(formula, &ERROR, &COMMENT, NULL);
	  if(number_mode == NUMBER_INT)
	    result = (int)result_f;

	  if(COMMENT || ERROR) continue;
	  //////////////////////////////////////////////
	  //output the result
	  //////////////////////////////////////////////
	  output_res(output_format, result_f, result);
    }
  FREE_AND_RETURN(file_content)
  return;
}