/* 
 *    Programmed By: Mohammed Isam Mohammed [mohammed_isam1984@yahoo.com]
 *    Copyright 2014, 2015 (c)
 * 
 *    file: input_free.c
 *    This file is part of tcalc.
 *
 *    tcalc is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    tcalc is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with tcalc.  If not, see <http://www.gnu.org/licenses/>.
 */    

#include "defs.h"

#define ERROR_CONTINUE(s)	{fprintf(stderr, s); continue;}

char buf[128];

/*
 * Will open the file passed in the *formula string
 * and perform requested operation:
 * 0 = read and process (input command)
 * 1 = list file contents only
 * 2 = list functions defined in the file
 */
void __input_file(char *formula, int request)
{
  int l = 0;
  if(request == 0) l = strstr(formula, "input")-formula+5;
  else if(request == 1) l = strstr(formula, "list")-formula+4;
  else if(request == 2) l = strstr(formula, "listfuncs")-formula+9;
  
  while(is_whitespace(formula[l])) l++;
  /* remove the trailing newline */
  int len = strlen(formula)-1;
  if(formula[len] == '\n') formula[len] = '\0';
  if(formula[l] == '\0')
  {
    fprintf(stderr, "Error: Missing file name\n");
    return;  
  }
  FILE *file;
  if((file = fopen(formula+l, "rb")) == NULL)
  {
    fprintf(stderr, "Error opening file: %s\n", formula+l);
    return;
  }
  if(!QUIET) printf("Reading from %s\n", formula+l);
  input_from_file_mode(file, request);
  fclose(file);
}

/*
 * Change number input format.
 */
int change_number_format(char *formula)
{
  char *f = strchr(formula, '=');
  if(!f)
  {
    if(input_format == DECIMAL_FORMAT) printf("decimal\n");
    else if(input_format == HEX_FORMAT) printf("hex\n");
    else if(input_format == BINARY_FORMAT) printf("binary\n");
    else if(input_format == OCTAL_FORMAT) printf("octal\n");
    else printf("unknown\n");
    return 0;
  }
  f++;
  while(is_whitespace(f[0])) f++;
  if(f[strlen(f)-1] == '\n') f[strlen(f)-1] = '\0';
  if(f[0] == '\0')
  {
    printf("Error: Missing argument\n");
    return 0;
  }
  
  if(strcmp(f, "d") == 0 || strcmp(f, "decimal") == 0)
    input_format = DECIMAL_FORMAT;
  else if(strcmp(f, "x") == 0 || strcmp(f, "hex") == 0)
    input_format = HEX_FORMAT;
  else if(strcmp(f, "b") == 0 || strcmp(f, "binary") == 0)
    input_format = BINARY_FORMAT;
  else if(strcmp(f, "o") == 0 || strcmp(f, "octal") == 0)
    input_format = OCTAL_FORMAT;
  else
  {
    printf("Error: Unknown number format: %s\n", f);
  }
  return 1;
}

/*
 * Change number output format.
 */
int change_output_format(char *formula)
{
  char *f = strchr(formula, '=');
  if(!f)
  {
    if(output_format == DECIMAL_FORMAT) printf("decimal\n");
    else if(output_format == HEX_FORMAT) printf("hex\n");
    else if(output_format == BINARY_FORMAT) printf("binary\n");
    else if(output_format == OCTAL_FORMAT) printf("octal\n");
    else printf("all\n");
    return 0;
  }
  f++;
  while(is_whitespace(f[0])) f++;
  if(f[strlen(f)-1] == '\n') f[strlen(f)-1] = '\0';
  if(f[0] == '\0')
  {
    printf("Error: Missing argument\n");
    return 0;
  }
  
  if(strcmp(f, "d") == 0 || strcmp(f, "decimal") == 0)
    output_format = DECIMAL_FORMAT;
  else if(strcmp(f, "x") == 0 || strcmp(f, "hex") == 0)
    output_format = HEX_FORMAT;
  else if(strcmp(f, "b") == 0 || strcmp(f, "binary") == 0)
    output_format = BINARY_FORMAT;
  else if(strcmp(f, "o") == 0 || strcmp(f, "octal") == 0)
    output_format = OCTAL_FORMAT;
  else if(strcmp(f, "a") == 0 || strcmp(f, "all") == 0)
    output_format = ALL_FORMATS;
  else
  {
    printf("Error: Unknown number format: %s\n", f);
  }
  return 1;
}

/*
 * Convert given number to the defined output format.
 */
void input_convert_numbers(char *formula)
{
  int j;
  for(j = 7; j < strlen(formula); j++)
    if(formula[j] != ' ' && formula[j] != '\n' && formula[j] != '\t')
    { convert_numbers(formula+j); break; }
}

/*
 * Called from input_do_factorial() to extract a numerical
 * from a string. We don't use normal libC functions for that
 * because our numbers can be Hex or Binary also!.
 */
int __get_number(char *n, int *res)
{
  int i = 0;
  double result_f = 0;
  char ERROR = 0;
  if(n[0] == '0' && (n[1] == 'x' || n[1] == 'X')) i = 2;
  else if(n[0] == 'b' || n[0] == 'B') i = 1;
  
  while(is_hex_digit(n[i])) i++;
  strncpy(buf, n, i);
  buf[i] = '\0';
  do_op(buf, '.', &result_f, &ERROR);
  if(ERROR) return 0;
  *res = (int)result_f;
  return i;
}

/*
 * Called from input_do_factorial() to do the actual
 * factorial calculation.
 */
unsigned long __do_factorial(unsigned long n)
{
  if(n == 1) return 1;
  else return n * __do_factorial(n-1);
}

/*
 * Called from input_free_mode() to do number factorial.
 * Does some checking before calling __do_factorial() to
 * do the actual work.
 */
unsigned long input_do_factorial(char *formula)
{
  int j;
  for(j = 10; j < strlen(formula); j++)
    if(formula[j] != ' ' && formula[j] != '\n' && formula[j] != '\t')
    {
      int res = 0, i = 0;
      i = __get_number(formula+j, &res);
      if(i && res != 0)
	return __do_factorial((unsigned long)res);
      else break;
    }
  return 0;
}

/*
 * Checks the passed order to print a multiplication table,
 * before calling print_mult_table() to actually print out
 * the table.
 */
void input_mult_table(char *formula)
{
  int j, i, k;
  int from, to, step;
  int index = -1;
  int indices[3] = {1, 12, 1};
  for(j = 5; j < strlen(formula); j++)
    if(formula[j] != ' ' && formula[j] != '\n' && formula[j] != '\t')
    {
      if(strncmp("from", formula+j, 4) == 0) { index = 0; j += 4; }
      else if(strncmp("to", formula+j, 2) == 0) { index = 1; j += 2; }
      else if(strncmp("step", formula+j, 4) == 0) { index = 2; j += 4; }
      else if(formula[j] >= '0' && formula[j] <= '9')
      {
	i = __get_number(formula+j, &k);
	if(!i) break;
	if(index >= 0) { indices[index] = k; index = -1; }
	else
	{
	  from = indices[0]; to = indices[1]; step = indices[2];
	  print_mult_table(formula+j, from, to, step);
	  return;
	}
	j += i;
      }
      else break;
    }
  /* if we reached here, there is an error! */
  printf("Error in the 'table' directive\n");
}



/*
 * The big boss who processes user input from the terminal
 * and then calls different functions accordingly.
 */
void input_free_mode()
{
    char formula[1024];
    int start = 0;
    while(1)
    {
	  formula[1023] = '\0';
	  SCOPE = GLOBAL_SCOPE;
	  printf(">");
	  char *s = fgets(formula+start, 1023, stdin);
	  if(!s) continue;
	  if(formula[strlen(formula)-2] == '\\')
	  {
	    start = strlen(formula)-2;
	    continue;
	  }
	  start = 0;
	  s = trim(formula, 1);
	  strcpy(formula, s);
	  if(formula[0] == '\n') continue;
	  if(strcmp(formula, "end\n") == 0) break;
	  
	  if(formula[0] == '#') continue;

	  int p = proc_input(formula, NULL);
	  if(p == 6) break;		/* means 'end' */
	  else if(p >= 0) continue;	/* means everything was processed right */
	  /* NOTE: negative return value from proc_input() means it couldn't
	   * process the directive. See if we can do it here.*/
	  
	  if(strncmp(formula, "func", 4) == 0)
	  {
	    int l = 4;
	    while(is_whitespace(formula[l])) l++;
	    if(l == 4 || formula[l] < 'a' || formula[l] > 'z')
	    {
	      printf("Error: Missing or Invalid function name\n");
	      continue;
	    }
	    int l2 = l+1;
	    int len = strlen(formula);
	    while(l2 < len && formula[l2] >= 'a' && formula[l2] <= 'z') l2++;
	    
	    int j;
	    
	    if(is_reserved_func(formula+l, l2-l)) continue;
	    if(is_custom_func(formula+l, l2-l)) continue;
	    struct custom_func_s *new_func = create_custom_func(formula+l, l2-l, 0);
	    if(!new_func) continue;
	    l = 0;
	    int l3 = 0;
	    for(j = l2; j < len; j++)
	    {
	      if(formula[j] == '(') l = j+1;
	      if(formula[j] == ')') l3 = j-1;
	    }
	    if(l)
	    {
	      if(!l3)
	      {
		printf("Error: Missing ')'\n");
		continue;
	      }
	      if(!add_function_args(new_func, formula+l, l3-l+1))
	      {
		delete_custom_func(new_func);
		ERROR_CONTINUE("Error: Insufficient memory to save function args\n")
	      }
	    }
	    new_func->func_body = (char *)malloc(FUNCTION_BODY_BLOCK_SIZE);
	    if(!new_func->func_body)
	    {
	      delete_custom_func(new_func);
	      ERROR_CONTINUE("Error: Insufficient memory to save function body\n")
	    }
	    new_func->func_body[0] = '\0';
	    new_func->func_body_res = FUNCTION_BODY_BLOCK_SIZE;
	    while(1)
	    {
	      printf("->");
	      s = fgets(formula, 1023, stdin);
	      if(!s) continue;
	      s = trim(formula, 1);
	      strcpy(formula, s);
	      if(strcmp(formula, "q\n") == 0 || strcmp(formula, "quit\n") == 0)
	      {
		delete_custom_func(new_func);
		break;
	      }
	      new_func->func_body_len += strlen(formula)+1;
	      if(new_func->func_body_len >= new_func->func_body_res)
	      {
		new_func->func_body_res = (int)(new_func->func_body_len/FUNCTION_BODY_BLOCK_SIZE)*FUNCTION_BODY_BLOCK_SIZE;
		if(new_func->func_body_len%FUNCTION_BODY_BLOCK_SIZE)
		  new_func->func_body_res += FUNCTION_BODY_BLOCK_SIZE;
		new_func->func_body = (char *)realloc(new_func->func_body,
						      new_func->func_body_res);
		if(!new_func->func_body)
		{
		  printf("Error: Insufficient memory to save function body\n");
		  delete_custom_func(new_func);
		  break;
		}
	      }
	      strcat(new_func->func_body, formula);
	      if(strchr(formula, '}')) break;
	    }
	    continue;
	  }
	  //////////////////////////////////////
	  //////////////////////////////////////
	  
	  char COMMENT = 0;
	  char ERROR = 0;
	  long result = 0;
	  double result_f = 0;
	  result_f = calc(formula, &ERROR, &COMMENT, NULL);
	  if(number_mode == NUMBER_INT)
	    result = (int)result_f;

	  //////////////////////////////////////////////
	  //output the result
	  //////////////////////////////////////////////
	  if(ERROR) continue;
	  
	  output_res(output_format, result_f, result);
    }
    return;
}

/*
 * Print the help menu.
 */
void show_help()
{
  printf("Commands [followed by ENTER]:\n");
  printf("convert n           Convert n to the output format specified by 'output' below\n");
  printf("decimal=n           How many digits are shown right to decimal point\n");
  printf("echo n              Print (or echo back) the string, n\n");
  printf("format=n            Input format [binary (b), hex (x), decimal (d), octal(o)]\n");
  printf("factorial n         Print the factorial of number n\n");
  printf("func                Define a custom function\n");
  printf("funcs               Print a list of custom-defined functions\n");
  printf("h, help             Show help\n");
  printf("input filename      Read & execute a formula file\n");
  printf("list filename       List (only) contents of a formula file\n");
  printf("listfuncs filename  List (only) formulae defined in a formula file\n");
  printf("math                Print a list of math functions\n");
  printf("output=n            Output format [binary (b), hex (x), decimal (d), octal(o),\n"
	 "                    all formats (a)\n");
  printf("q, quit             Quit\n");
  printf("table [from] [to] [step] n\n"
         "                    Print the multiplication table of number 'n', table ranges\n"
	 "                    between 'from' and 'to', with a stepping of 'step'\n");
  printf("undef func n        Undefine previously defined custom function, n\n");
  printf("undef n             Undefine previously defined variable, n\n");
  printf("var n [= value]     Define a variable, and (possibly) initialize it's value\n");
  printf("vars                Print a list of defined variables\n");
  printf("ver, version        Show version\n");
}
