/* 
 *    Programmed By: Mohammed Isam Mohammed [mohammed_isam1984@yahoo.com]
 *    Copyright 2014, 2015, 2016 (c)
 * 
 *    file: numbers.c
 *    This file is part of tcalc.
 *
 *    tcalc is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    tcalc is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with tcalc.  If not, see <http://www.gnu.org/licenses/>.
 */    

#include "defs.h"

extern char res2[64];	/* calc.c */

char hex_chars[] =
{
  '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
  'A', 'B', 'C', 'D', 'E', 'F'
};

char is_op(char op)
{
  if(op == '+' || op == '-' ||
     op == '/' || op == '*' ||
     op == '%' || op == '^' ||
     op == '(' || op == ')' ||
     op == '\n' || op == '\t' ||
     op == '\0' || op == ' ' ||
     op == '&' || op == '|')
    return 1;
  return 0;
}

char is_hex_digit(char hex_digit)
{
  int i;
  for(i = 0; i < 16; i++)
    if(hex_digit == hex_chars[i] || hex_digit == hex_chars[i]+32)
      return 1;
  return 0;
}

/*
 * Make sure the passed hex number is a valid hex.
 */
int authenticate_hex(char *hex, char *error)
{
  *error = 0;
  int len = strlen(hex);
  if(!hex || len < 3) goto hex_error;
  /* First make sure it is hex! */
  if(strstr(hex, "0X") != hex && strstr(hex, "0x") != hex) goto hex_error;
  int j = 2;
  /* Then get rid of any trailing non-hex figures */
  while(j < len)
  {
    char found = is_hex_digit(hex[j]);
    if(found) { j++; continue; }
    break;
  }
  len = j;
  if(!is_op(hex[len])) goto hex_error;
  return len;
hex_error:
  *error = 1;
  return 0;
}

/*
 * Make sure the passed octal number is a valid octal.
 */
int authenticate_octal(char *octal, char *error)
{
  *error = 0;
  int j = 0;
  if(!octal) goto octal_error;
  int len = strlen(octal);
  /* First get rid of any trailing non-numbers */
  while(j < len)
  {
    if(octal[j] < '0' || octal[j] > '7') break;
    j++;
  }
  len = j;
  if(!is_op(octal[len])) goto octal_error;
  return len;
octal_error:
  *error = 1;
  return 0;
}

/*
 * Make sure the passed binary number is a valid binary.
 */
int authenticate_binary(char *binary, char *error)
{
  int j = 1;
  *error = 0;
  /* Make sure it is binary! */
  if(!binary) goto binary_error;
  //if(binary[0] != 'b') goto binary_error;
  int len = strlen(binary);
  /* Get rid of any trailing non-numbers */
  while(j < len)
  {
    if(binary[j] != '0' && binary[j] != '1') break;
    j++;
  }
  len = j;
  if(!is_op(binary[len])) goto binary_error;
  return len;
binary_error:
  *error = 1;
  return 0;
}

unsigned long hex_to_decimal(char *hex, char *error)
{
  int i, j, len = authenticate_hex(hex, error);
  if(*error) goto hex_error;
  
  unsigned long res = 0;
  for(j = 2; j < len; j++)
  {
    char found = 0;
    for(i = 0; i < 16; i++)
      if(hex[j] == hex_chars[i] || hex[j] == hex_chars[i]+32)
      { res = (res*16) + i; found = 1; break; }
    if(!found) goto hex_error;
  }
  return res;
  
hex_error:
  printf("Error: Invalid hexadecimal input\n");
  *error = 1; return 0;
}


unsigned long binary_to_decimal(char *binary, char *error)
{
  unsigned long res = 0;
  int i, j = 1;

  int len = authenticate_octal(binary, error);
  if(*error) goto binary_error;
  j = len;

  j--;
  for(i = 0; i < len; i++)
  { res += (((binary[i] == '1') ? 1 : 0) << j); j--; }
  return res;
  
binary_error:
  printf("Error: Invalid binary input\n");
  *error = 1; return 0;
}


unsigned long octal_to_decimal(char *octal, char *error)
{
  unsigned long res = 0;
  int i, j = 0;
  int len = authenticate_octal(octal, error);
  if(*error) goto octal_error;

  /* And make sure it is octal! */
  if(octal[0] != '0') goto octal_error;

  for(j = 0; j < len; j++)
  {
    if(octal[j] > '7' || octal[j] < '0') break;
    i = octal[j]-48; /* ASCII for '0' */
    res = (res*8) + i;
  }
  return res;

octal_error:
  printf("Error: Invalid octal input\n");
  *error = 1; return 0;
}


void decimal_to_binary(unsigned long decimal, char **res, char *error)
{
  *error = 0;
  char res2[64];
  res2[63] = '\0';
  int j = 62, i;
  long quotient = 1, remainder = 0;
  while(quotient)
  {
    quotient = decimal/2;
    remainder = decimal%2;
    res2[j] = remainder ? '1' : '0';
    j--;
    decimal = quotient;
  }
  int last = 62-j;
  for(i = 0; i < last; i++)
  {
    (*res)[i] = res2[++j];
  }
  (*res)[i] = '\0';
}


void decimal_to_hex(unsigned long decimal, char **res, char *error)
{
  *error = 0;
  char res2[64];
  sprintf(res2, "%lx", decimal);
  int pos = 2, i = 0, j = strlen(res2);
  if(j % 2) i = 1;
  (*res)[0] = '0';
  (*res)[1] = 'x';
  if(i) { (*res)[2] = '0'; pos++; }
  j = 0;
  for(i = 0; i < strlen(res2); i++)
    (*res)[pos++] = res2[j++];
  (*res)[pos] = '\0';
}


void convert_numbers(char *input)
{
  if(input == NULL) return;
  unsigned long res;
  char error;
  char *res3 = res2;
  int len = strlen(input);
  if(input[0] == 'b' || input[0] == 'B')
  {
    res = binary_to_decimal(input+1, &error);
    if(error) return;
    if(output_format == DECIMAL_FORMAT)
      printf("%+ld (%lu)\n", res, res);
    else if(output_format == OCTAL_FORMAT)
      printf("0%lo\n", res);
    else if(output_format == HEX_FORMAT)
    {
      decimal_to_hex(res, &res3, &error);
      if(*res2) { printf("%s\n", res2); }
    }
    else if(output_format == BINARY_FORMAT)
    {
      if(input[len-1] == '\n') printf("%s", input);
      else                     printf("%s\n", input);
    }
    else if(output_format == ALL_FORMATS)
    {
      printf("%+ld (%lu)\n", res, res);
      printf("0%lo\n", res);
      decimal_to_hex(res, &res3, &error);
      if(*res2) { printf("%s\n", res2); }
      if(input[len-1] == '\n') printf("%s", input);
      else                     printf("%s\n", input);
    }
  }
  else if((input[0] == '0' && input[1] == 'X') ||
	  (input[0] == '0' && input[1] == 'x'))
  {
    res = hex_to_decimal(input, &error);
    if(error) return;
    //printf("%s(): res = %ld (%lu), input = '%s'\n", __func__, res, res, input);
    if(output_format == DECIMAL_FORMAT)
      printf("%+ld (%lu)\n", res, res);
    else if(output_format == OCTAL_FORMAT)
      printf("0%lo\n", res);
    else if(output_format == HEX_FORMAT)
    {
      if(input[len-1] == '\n') printf("%s", input);
      else                     printf("%s\n", input);
    }
    else if(output_format == BINARY_FORMAT)
    {
      decimal_to_binary(res, &res3, &error);
      if(*res2) { printf("b%s\n", res2); }
    }
    else if(output_format == ALL_FORMATS)
    {
      printf("%+ld (%lu)\n", res, res);
      printf("0%lo\n", res);
      if(input[len-1] == '\n') printf("%s", input);
      else                     printf("%s\n", input);
      decimal_to_binary(res, &res3, &error);
      if(*res2) { printf("b%s\n", res2); }
    }
  }
  else if(input[0] == '0')
  {
    res = octal_to_decimal(input, &error);
    if(error) return;
    if(output_format == DECIMAL_FORMAT)
      printf("%+ld (%lu)\n", res, res);
    else if(output_format == OCTAL_FORMAT)
    {
      if(input[len-1] == '\n') printf("%s", input);
      else                     printf("%s\n", input);
    }
    else if(output_format == HEX_FORMAT)
    {
      decimal_to_hex(res, &res3, &error);
      if(*res2) { printf("%s\n", res2); }
    }
    else if(output_format == BINARY_FORMAT)
    {
      decimal_to_binary(res, &res3, &error);
      if(*res2) { printf("b%s\n", res2); }
    }
    else if(output_format == ALL_FORMATS)
    {
      printf("%+ld (%lu)\n", res, res);
      if(input[len-1] == '\n') printf("%s", input);
      else                     printf("%s\n", input);
      decimal_to_hex(res, &res3, &error);
      if(*res2) { printf("%s\n", res2); }
      decimal_to_binary(res, &res3, &error);
      if(*res2) { printf("b%s\n", res2); }
    }
  }
  else
  {
    res = atol(input);
    if(output_format == DECIMAL_FORMAT)
      printf("%+ld (%lu)\n", res, res);
    else if(output_format == OCTAL_FORMAT)
      printf("0%lo\n", res);
    else if(output_format == HEX_FORMAT)
    {
      decimal_to_hex(res, &res3, &error);
      if(*res2) { printf("%s\n", res2); }
    }
    else if(output_format == BINARY_FORMAT)
    {
      decimal_to_binary(res, &res3, &error);
      if(*res2) { printf("b%s\n", res2); }
    }
    else if(output_format == ALL_FORMATS)
    {
      printf("%+ld (%lu)\n", res, res);
      printf("0%o\n", (unsigned)res);
      decimal_to_hex(res, &res3, &error);
      if(*res2) { printf("%s\n", res2); }
      decimal_to_binary(res, &res3, &error);
      if(*res2) { printf("b%s\n", res2); }
    }
  }
}
