/* 
 *    Programmed By: Mohammed Isam Mohammed [mohammed_isam1984@yahoo.com]
 *    Copyright 2014, 2015 (c)
 * 
 *    file: proc_input.c
 *    This file is part of tcalc.
 *
 *    tcalc is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    tcalc is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with tcalc.  If not, see <http://www.gnu.org/licenses/>.
 */    

#include "defs.h"

int next_whitespace(char *str)
{
  char *orig_str = str;
  while(*str && !is_whitespace(*str) && *str != '=') str++;
  return str-orig_str;
}

int proc_input(char *formula, struct custom_func_s *active_func)
{
  /* isolate the keyword from the rest of the formula line ... */
  int whitespace = next_whitespace(formula);
  if(whitespace > 31)
  {
    printf("Error: Very long command word indeed!\n");
    return 7;
  }
  char f[32];
  strncpy(f, formula, whitespace);
  f[whitespace] = '\0';
  /* ... then compare looking for a hit */
  if(strcmp(f, "ver") == 0 || strcmp(f, "version") == 0)
  {
    printf("Version %s\n", tcalc_ver);
  }
  else if(strcmp(f, "h") == 0 || strcmp(f, "help") == 0)
  {
    show_help();
  }
  else if(strcmp(f, "math") == 0)
  {
    print_func_descriptions();
  }
  else if(strcmp(f, "funcs") == 0)
  {
    print_custom_funcs();
  }
  else if(strcmp(f, "decimal") == 0)
  {
    char *d = strchr(formula, '=');
    if(!d)
    {
      printf("%d\n", decimal_point);
      return 0;
    }
    d++;
    while(is_whitespace(d[0])) d++;
    if(d[0] == '\0')
    {
      printf("Error: Missing argument\n");
      return 3;
    }
    decimal_point = atoi(d);
    if(decimal_point > MAX_DECIMAL_POINT) decimal_point = MAX_DECIMAL_POINT;
  }
  else if(strcmp(f, "format") == 0)
  {
    if(!change_number_format(formula)) return 1;
  }
  else if(strcmp(f, "output") == 0)
  {
    if(!change_output_format(formula)) return 2;
  }
  else if(strcmp(f, "convert") == 0)
  {
    input_convert_numbers(formula);
  }
  else if(strcmp(f, "table") == 0)
  {
    input_mult_table(formula);
  }
  else if(strcmp(f, "factorial") == 0)
  {
    long fact = input_do_factorial(formula);
    if(!fact)
    {
      printf("Error: Invalid or missing argument\n");
      return 4;
    }
    else
    {
      number_mode = NUMBER_INT;
      output_res(output_format, 0.0, fact);
    }
  }
  else if(strcmp(f, "input") == 0)
  {
    __input_file(formula, 0);
  }
  else if(strcmp(f, "listfuncs") == 0)
  {
    __input_file(formula, 2);
  }
  else if(strcmp(f, "list") == 0)
  {
    __input_file(formula, 1);
  }
  else if(strcmp(f, "undef") == 0)
  {
    int k = 5;
    while(is_whitespace(formula[k])) k++;
    if(k == 5 || formula[k] == '\0')
    {
undef_return_err:
      printf("Error in 'undef' directive\n");
      return 5;
    }
    remove_trailing_newline(formula);
    /* 'undef func' directive */
    if(strncmp(formula+k, "func", 4) == 0)
    {
      k += 4;
      int old_k = k;
      while(is_whitespace(formula[k])) k++;
      if(old_k == k) goto undef_return_err;
      delete_custom_func_by_name(formula+k);
    }
    /* 'undef' variable directive */
    else
    {
      /*
       * TODO: Correct this.
       */
      //printf("Variable %s is not defined\n", formula+k);
      delete_var_by_name(formula+k, active_func);
    }
  }
  else if(strcmp(f, "vars") == 0)
  {
    print_vars(active_func);
  }
  else if(strcmp(f, "var") == 0)
  {
    //printf("proc_input()\n");
    char *tmp2 = (formula+whitespace);
    whitespace = 0;
    while(is_whitespace(tmp2[whitespace])) whitespace++;
    tmp2 += whitespace;
    whitespace = next_whitespace(tmp2);
    if(whitespace > MAX_VARIABLE_LEN)
    {
      printf("Error: Very long variable name (Max. is %d)\n", MAX_VARIABLE_LEN);
      return 8;
    }
    if(tmp2[0] == '\0')
    {
      printf("Error: Missing variable name\n");
      return 8;
    }
    remove_trailing_newline(formula);
    strncpy(f, tmp2, whitespace);
    f[whitespace] = '\0';

    char *v = strchr(formula+whitespace, '=');
    double val = 0;
    if(v)
    {
      v++;
      while(is_whitespace(v[0])) v++;
      if(v[0] == '\0')
      {
var_val_error:
	printf("Error: Missing or Invalid RValue\n");
	return 9;
      }
      char ERROR;
      do_op(v, '.', &val, &ERROR);
      if(ERROR) goto var_val_error;
    }
    
    int index = 0;
    
    if(active_func == NULL)		/* global scope */
    {
      if(total_global_vars >= MAX_GLOBAL_VARS)
      {
	printf("Error: maximum number of global variables reached.\n");
	return 10;
      }
      /* WARNING: We will overwrite variables without warning! */
      for(index = 0; index < total_global_vars; index++)
	if(strcmp(global_var_name[index], f) == 0) break;
      strcpy(global_var_name[index], f);
      global_var_value[index] = val;
      if(index == total_global_vars) total_global_vars++;
    }
    else				/* function scope */
    {
      if(active_func->total_vars >= MAX_VARIABLES)
      {
	printf("Error: maximum number of local function variables reached.\n");
	return 10;
      }
      /* WARNING: We will overwrite variables without warning! */
      for(index = 0; index < active_func->total_vars; index++)
	if(strcmp(active_func->var_name[index], f) == 0) break;
      strcpy(active_func->var_name[index], f);
      active_func->var_value[index] = val;
      if(index == active_func->total_vars) active_func->total_vars++;
      //print_vars(active_func);
    }
  }
  else if(strcmp(f, "echo") == 0)
  {
    char *tmp2 = (formula+whitespace);
    whitespace = 0;
    while(is_whitespace(tmp2[whitespace])) whitespace++;
    tmp2 += whitespace;
    if(tmp2[0] == '"') tmp2++;
    int len = strlen(tmp2);
    if(tmp2[len-1] == '"')
    {
      tmp2[len-1] = tmp2[len  ]; len--;
    }
    else if(tmp2[len-2] == '"')
    {
      tmp2[len-2] = tmp2[len-1];
      tmp2[len-1] = tmp2[len  ]; len--;
    }
    if(tmp2[len-1] == '\n')
      printf("%s", tmp2);
    else
      printf("%s\n", tmp2);
  }
  else if(strcmp(f, "q") == 0 || strcmp(f, "Q") == 0 ||
	  strcmp(f, "quit") == 0 || strcmp(f, "end") == 0)
  {
    return 6;
  }
  //////////////////////////////
  //////////////////////////////
  else
  {
    return -1;
  }
  return 0;
}
